@@ -1,5 +1,5 @@
 package Form::Tiny::Path;
-$Form::Tiny::Path::VERSION = '2.20';
+$Form::Tiny::Path::VERSION = '2.21';
 use v5.10;
 use strict;
 use warnings;
@@ -28,6 +28,19 @@ has 'meta' => (
 	required => 1,
 );
 
+# cache for meta array positions
+has 'meta_arrays' => (
+	is => 'ro',
+	default => sub {
+		return [
+			map {
+				$_ eq 'ARRAY'
+			} @{$_[0]->meta}
+		];
+	},
+	init_arg => undef,
+);
+
 sub BUILD
 {
 	my ($self) = @_;
@@ -41,7 +54,7 @@ sub BUILD
 			if scalar grep { length $_ eq 0 } @parts;
 
 		croak 'path specified started with an array: ' . $self->dump
-			if $self->meta->[0] eq 'ARRAY';
+			if $self->meta_arrays->[0];
 	}
 }
 
@@ -111,6 +124,15 @@ sub append
 	return $self;
 }
 
+sub append_path
+{
+	my ($self, $other_path) = @_;
+
+	push @{$self->path}, @{$other_path->path};
+	push @{$self->meta}, @{$other_path->meta};
+	return $self;
+}
+
 sub make_name_path
 {
 	my ($self, $prefix) = @_;
@@ -146,11 +168,11 @@ sub follow
 
 	my @found = ($structure);
 	my @path = @{$self->path};
-	my @meta = @{$self->meta};
+	my $meta = $self->meta_arrays;
 	my $has_array = 0;
 
 	for my $ind (0 .. $#path) {
-		my $is_array = $meta[$ind] eq 'ARRAY';
+		my $is_array = $meta->[$ind];
 		my @new_found;
 
 		for my $item (@found) {
